
from generallibrary import deco_cache, typeChecker, getRows, initBases


class Path_Spreadsheet:
    """ Spreadsheet methods for Path. """
    @property
    @deco_cache()
    def spreadsheet(self):
        """ Easy access to a dynamically one-time created Spreadsheet class, initalized with self (Path). """
        return self._spreadsheet()(self)

    @staticmethod
    @deco_cache()
    def _spreadsheet():
        @initBases
        class _Spreadsheet(_Extension):
            """ Contains all functionality to handle spreadsheet files.
                Such as CSV and TSV. """

            import pandas
            import csv

            pd, csv = pandas, csv

            def __init__(self, path):
                self.path = path

            def write(self, df=None, overwrite=False):
                """ Can write a bunch of different DataFrames to a TSV file.
                    Doesn't support advanced pandas functionality.
                    Should work with: Keys, Index and Transposed (8 combinations)
                    If DataFrame has both keys and index then cell A1 becomes NaN

                    TODO: Make it order columns if there are any so that they line up with append.

                    :param pandas.DataFrame df: Serializable by JSON
                    :param overwrite: Whether to allow overwriting or not. """

                typeChecker(df, self.pd.DataFrame)

                with self.WriteContext(self.path, overwrite=overwrite) as write_path:
                    if df.empty:
                        write_path.write()
                        return False, False

                    useHeader = self._indexIsNamed(df.columns)
                    useIndex = self._indexIsNamed(df.index)

                    df.to_csv(str(write_path), sep="\t", header=useHeader, index=useIndex)

                    return useHeader, useIndex

            def read(self, header=False, column=False):
                """
                If any cell becomes NaN then the header and column parameters are overriden silently.

                Can read a bunch of different DataFrames to a TSV file.
                Doesn't support advanced pandas functionality.
                Should work with: Keys, Index, Transposed, Header, Column (32 combinations).
                DataFrame in file can have a NaN A1 cell.

                :param bool header: Use headers or not, overriden if any top left cell is NaN
                :param bool column: Use columns or not, overriden if any top left cell is NaN
                :rtype: pd.DataFrame
                """

                header = "infer" if header else None
                column = 0 if column else None

                with self.ReadContext(self.path) as read_path:
                    try:
                        df = self._read_helper(read_path, header, column)
                    except self.pd.errors.EmptyDataError:
                        return self.pd.DataFrame()

                    # Get rid of empty cell (Happens if file was written with header=True, column=True)
                    headerFalseColumnFalse = self.pd.isna(df.iat[0, 0])
                    headerFalseColumnTrue = self.pd.isna(df.index[0])
                    headerTrueColumnFalse = str(df.columns[0]).startswith("Unnamed: ")
                    if headerFalseColumnFalse or headerFalseColumnTrue or headerTrueColumnFalse:
                        header = "infer"
                        column = 0
                        df = self._read_helper(read_path, header, column)
                    else:
                        # Get rid of name in index (Happens if file doesn't have an index and column=True)
                        # Doesn't happen other way around for some reason, guess it's the internal order in pandas
                        if df.index.name is not None:
                            if header is None and column == 0:
                                df.index.rename(None, inplace=True)
                            else:
                                header = None
                                column = None
                                df = self._read_helper(read_path, header, column)

                    if not self._indexIsNamed(df.columns):
                        df.columns = self.pd.RangeIndex(len(df.columns))
                    if not self._indexIsNamed(df.index):
                        df.index = self.pd.RangeIndex(len(df.index))

                    return df.convert_dtypes()

            def append(self, obj):
                """
                Append an obj containing lists or dicts to the end of a TSV file.
                If a dict is given and there are iterables as values then the keys of the dict are the first value in each row.
                Otherwise keys in dicts are ignored.

                Identical append objects
                 | [[1, 2, 3], [4, 5, 6]]
                 | [{"a": 1, "b": 2, "c": 3}, {"d": 4, "e": 5, "f": 6}]
                 | {1: {"b": 2, "c": 3}, 4: {"e": 5, "f": 6}}
                 | {1: [2, 3], 4: [5, 6]}

                TODO: Should probably support DataFrame and Series as well.

                :param obj: Iterable (Optionally inside another iterable) or a value for a single cell
                """
                def _append_helper(stream):
                    writer = self.csv.writer(stream, delimiter="\t", lineterminator="\n")
                    for row in getRows(obj):
                        writer.writerow(row)

                with self.AppendContext(self.path) as append_path:
                    return append_path.open_operation("a", _append_helper)


            def _indexIsNamed(self, index):
                """
                Simple version to see if a DataFrame index is named or not
                :param index: DataFrame's index (columns or index)
                """
                if str(index[0]) == "0" or str(index[0]) == "1":
                    return False
                else:
                    return True

            def _read_helper(self, path, header, column):
                return self.pd.read_csv(str(path), sep="\t", header=header, index_col=column).convert_dtypes()

            def _append_helper(self, iterable_obj, key=None):
                """
                Takes an object and returns a list of rows to use for appending.

                :param iterable_obj: Iterable
                :param key: If iterableObj had a key to assigned it it's given here
                :return: A
                """
                row = [key] if key else []
                if isinstance(iterable_obj, (list, tuple)):
                    row.extend(iterable_obj)
                elif isinstance(iterable_obj, dict):
                    for _, value in sorted(iterable_obj.items()):
                        row.append(value)
                return row

        return _Spreadsheet


from generalfile.path_operations import _Extension









































"""Gather all custom errors here."""

class _MiddleGroundError(Exception): """Parent of all middle ground errors."""
class CaseSensitivityError(_MiddleGroundError): """Raised when an existing file matches a path but not the exact case."""
class InvalidCharacterError(_MiddleGroundError): """Raised when using an invalid character."""









import re


class Path_Strings:
    """ String operations for Path. """
    def __getitem__(self, item):
        """ Get character from path string.

            :param generalfile.Path self: """
        return self.Path(self.path.__getitem__(item))

    @property
    def get_part(self):
        """ Split path using it's delimiter. item=0 with an absolute path gives an empty string on a posix system.

            :param generalfile.Path self:
            :param i: Index of part"""
        return self.path.split(self.path_delimiter)

    def get_replaced_alternative_characters(self):
        """ Get a dictionary of all characters that are replaced for the alternative path.

            :param generalfile.Path self: """
        return {
            self.path_delimiter: "&#47;",
            ":": "&#58",
            ".": "&#46;"
        }

    def get_alternative_path(self):
        """ Get path using alternative delimiter and alternative root for windows.

            :param generalfile.Path self:
            :rtype: generalfile.Path """
        path = str(self.absolute())
        for char, alternative in self.get_replaced_alternative_characters().items():
            path = path.replace(char, alternative)
        return self.Path(path)

    def get_lock_path(self):
        """ Get absolute lock path pointing to actual lock.

            :param generalfile.Path self: """
        return self.get_lock_dir() / self.absolute().get_alternative_path()

    def get_path_from_alternative(self):
        """ Get path from an alternative representation with or without leading lock dir.

            :param generalfile.Path self:
            :rtype: generalfile.Path """

        path = str(self.remove_start(self.get_lock_dir()))
        for char, alternative in self.get_replaced_alternative_characters().items():
            path = path.replace(alternative, char)
        return self.Path(path)

    def absolute(self):
        """ Get new Path as absolute.

            :param generalfile.Path self:
            :rtype: generalfile.Path """
        if self.is_absolute():
            return self
        else:
            return self.get_working_dir() / self
            # return self.Path(self._path.absolute())

    def relative(self, base=None):
        """ Get new Path as relative.

            :param generalfile.Path self:
            :param base: Defaults to working dir. """
        if self.is_relative() and base is None:
            return self
        else:
            if base is None:
                base = self.get_working_dir()
            return self.Path() if self == base else self.Path(self._path.relative_to(str(base)))

    def is_absolute(self):
        """ Get whether this Path is absolute.

            :param generalfile.Path self: """
        return self._path.is_absolute()

    def is_relative(self):
        """ Get whether this Path is relative.

            :param generalfile.Path self: """
        return not self.is_absolute()

    def startswith(self, path):
        """ Get whether this Path starts with given string.

            :param generalfile.Path self:
            :param str or Path path:"""
        path = self.Path(path)
        return str(self).startswith(str(path))

    def endswith(self, path):
        """ Get whether this Path ends with given string.

            :param generalfile.Path self:
            :param str or Path path:"""
        path = self.Path(path)
        return str(self).endswith(str(path))

    def remove_start(self, path):
        """ Remove a string from the start of this Path.

            :param generalfile.Path self:
            :param str or Path path:"""
        path = self.Path(path)
        str_path = str(path)
        if not self.startswith(str_path):
            return self
        else:
            new_path = self.Path(str(self)[len(str_path):])
            if str(new_path).startswith(path.path_delimiter):
                return new_path[1:]
            else:
                return new_path

    def remove_end(self, path):
        """ Remove a string from the end of this Path.

            :param generalfile.Path self:
            :param str or Path path:"""
        path = self.Path(path)
        str_path = str(path)
        if not self.endswith(str_path):
            return self
        else:
            new_path = self.Path(str(self)[:-len(str_path)])
            if str(new_path).endswith(path.path_delimiter):
                return new_path[:-1]
            else:
                return new_path

    def same_destination(self, path):
        """ See if two paths point to the same destination.

            :param generalfile.Path self:
            :param str or Path path:"""
        path = self.Path(path)
        return self.absolute() == path.absolute()

    def parts(self):
        """ Split path using it's delimiter.
            With an absolute path the first index is an empty string on a posix system. <- Note sure about that anymore, might be /

            :param generalfile.Path self: """
        return str(self).split(self.path_delimiter)

    def name(self):
        """ Get name of Path which is stem + suffix.

            :param generalfile.Path self: """
        return self._path.name

    def with_name(self, name):
        """ Get a new Path with new name which is stem + suffix.

            :param name: Name.
            :param generalfile.Path self:
            :rtype: generalfile.Path """
        return self.Path(self._path.with_name(str(name)))

    def stem(self):
        """ Get stem which is name without last suffix.

            :param generalfile.Path self: """
        return self._path.stem

    def with_stem(self, stem):
        """ Get a new Path with new stem which is name without last suffix.

            :param stem: New stem.
            :param generalfile.Path self:
            :rtype: generalfile.Path """
        return self.Path(self.with_name(f"{stem}{self.suffix()}"))

    def true_stem(self):
        """ Get true stem which is name without any suffixes.

            :param generalfile.Path self: """
        return self._path.stem.split(".")[0]

    def with_true_stem(self, true_stem):
        """ Get a new Path with new stem which is name without any suffixes.

            :param true_stem: New true stem.
            :param generalfile.Path self:
            :rtype: generalfile.Path """
        return self.Path(self.with_name(f"{true_stem}{''.join(self.suffixes())}"))

    def suffix(self):
        """ Get suffix which is name without stem.

            :param generalfile.Path self: """
        return self._path.suffix

    def with_suffix(self, suffix, index=-1):
        """ Get a new Path with a new suffix at any index.
            Index is automatically clamped if it's outside index range.
            Set suffix to `None` to remove a suffix.

            :param generalfile.Path self:
            :param suffix: New suffix, can be `None`.
            :param index: Suffix index to alter.
            :rtype: generalfile.Path """

        suffixes = self.suffixes()

        try:
            suffixes[index]
        except IndexError:
            if index >= len(suffixes):
                if suffix is None:
                    del suffixes[-1]
                else:
                    suffixes.append(suffix)
            else:
                if suffix is None:
                    del suffixes[0]
                else:
                    suffixes.insert(0, suffix)
        else:
            if suffix is None:
                del suffixes[index]
            else:
                suffixes[index] = suffix

        return self.with_name(f"{self.true_stem()}{''.join(suffixes)}")

    def suffixes(self):
        """ Get every suffix as a list.

            :param generalfile.Path self: """
        return self._path.suffixes

    def with_suffixes(self, suffixes):
        """ Get a new Path with a new list of suffixes.

            :param list or tuple suffixes: New list of suffixes.
            :param generalfile.Path self:
            :rtype: generalfile.Path """
        return self.Path(self.with_name(f"{self.true_stem()}{''.join(suffixes)}"))

    def match(self, *lines):
        """ Get whether any parts of this Path matches any given filter line.

            :param generalfile.Path self: """
        for part in self.parts():
            for line in lines:
                line = re.escape(line)
                line = line.replace(r"\*", ".+")
                pattern = f"^{line}$"
                if re.match(pattern, part, re.IGNORECASE):
                    return True
        return False
































from generallibrary import deco_cache, initBases


class Path_Text:  # Import in path.py
    """ Text methods for Path. """
    @property
    @deco_cache()
    def text(self):
        """ Easy access to a dynamically one-time created Text class, initalized with self (Path). """
        return self._text()(self)

    @staticmethod
    @deco_cache()
    def _text():
        @initBases
        class _Text(_Extension):
            def write(self, string=None, overwrite=False):
                """ Write to this path with a given string. """
                with self.WriteContext(self.path, overwrite=overwrite) as write_path:
                    return write_path.open_operation("w", lambda stream: stream.write(string))

            def read(self):
                """ Read from this path to get a string. """
                with self.ReadContext(self.path) as read_path:
                    return read_path.open_operation("r", lambda stream: stream.read())

            def append(self, string):
                """ Append to this path with a given string. """
                with self.AppendContext(self.path) as append_path:
                    return append_path.open_operation("a", lambda stream: stream.write(string))

        return _Text


from generalfile.path_operations import _Extension









































import pathlib
import appdirs
import os
import shutil
from send2trash import send2trash
import json
from distutils.dir_util import copy_tree
import time

from generallibrary import deco_cache

from generalfile.errors import CaseSensitivityError
from generalfile.decorators import deco_require_state, deco_preserve_working_dir, deco_return_if_removed



class _Context:
    def __init__(self, path):
        self.path = path
        self.temp_path = self.path.with_suffix(".tmp")
        self.lock = self.path.lock()

    def __enter__(self):
        self.lock.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.lock.__exit__(exc_type, exc_val, exc_tb)


class WriteContext(_Context):
    """ A context manager used for every write method. """
    def __init__(self, path, overwrite=False):
        super().__init__(path)
        self.overwrite = overwrite

    def __enter__(self):
        """ :rtype: generalfile.Path """
        if not self.overwrite and self.path.exists():
            raise FileExistsError(f"Path '{self.path}' already exists and overwrite is 'False'.")
        super().__enter__()
        self.path.get_parent().create_folder()
        return self.temp_path

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.temp_path.rename(self.path.name(), overwrite=True)
        super().__exit__(exc_type, exc_val, exc_tb)


class ReadContext(_Context):
    """ A context manager used for every read method. """
    def __init__(self, path):
        super().__init__(path)

    def __enter__(self):
        super().__enter__()
        return self.path

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)


class AppendContext(_Context):
    """ A context manager used for every append method. """
    def __init__(self, path):
        super().__init__(path)

    def __enter__(self):
        super().__enter__()
        if self.path.exists():
            self.path.copy(self.temp_path)
        return self.temp_path

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.temp_path.rename(self.path.name(), overwrite=True)
        super().__exit__(exc_type, exc_val, exc_tb)


class _Extension:
    WriteContext = WriteContext
    ReadContext = ReadContext
    AppendContext = AppendContext

    def __init__(self, path):
        self.path = path


class Path_Operations:
    """ File operations methods for Path. """
    _suffixIO = {"plain_text": ("txt", "md", ""), "spreadsheet": ("tsv", "csv")}
    timeout_seconds = 5
    dead_lock_seconds = 3
    _working_dir = None

    def open_operation(self, mode, func):
        """ Handles all open() calls. """
        with open(str(self), mode, encoding="utf-8") as stream:
            return func(stream)

        # Couldn't do this as it only fails when writing, not reading
        # try:
        #     with open(str(self), mode) as stream:
        #         return func(stream)
        # except UnicodeEncodeError:
        #     with open(str(self), mode, encoding="utf-8") as stream:
        #         return func(stream)


    def write(self, content=None, overwrite=False, indent=None):
        """ Write to this Path with JSON.

            :param generalfile.Path self:
            :param any content: Serializable by JSON
            :param overwrite: Whether to allow overwriting or not.
            :param indent: """
        content_json = json.dumps(content, indent=indent)
        with WriteContext(self, overwrite=overwrite) as write_path:
            write_path.open_operation("w", lambda stream: stream.write(content_json))
        return content_json

    def read(self):
        """ Read this Path with JSON.

            :param generalfile.Path self: """
        with ReadContext(self) as read_path:
            return read_path.open_operation("r", lambda stream: json.loads(stream.read()))

    @deco_require_state(exists=True)
    def rename(self, name=None, stem=None, suffix=None, overwrite=False):
        """ Rename this single file or folder to anything.

            TODO: Can we not just change signature to rename(self, new_path, overwrite=False) ?
                Then just require parents to be same and just use path's with_* methods.

            :param generalfile.Path self:
            :param name:
            :param stem:
            :param suffix:
            :param overwrite: """
        new_path = self
        for key, value in {"stem": stem, "suffix": suffix, "name": name}.items():
            if value is not None:
                new_path = getattr(new_path, f"with_{key}")(value)
        if new_path == self:
            return

        with self.lock(new_path):
            if overwrite:
                self._path.replace(str(new_path))
            else:
                self._path.rename(str(new_path))

    @deco_require_state(exists=True)
    def copy(self, new_path, overwrite=False):
        """ Copy a file or folder next to itself with a new name.
        If target exists then it is removed first, so it cannot add to existing folders, use `copy_to_folder` for that.

        :param generalfile.Path self:
        :param generalfile.Path or str new_path:
        :param overwrite:
        :return:
        """
        new_path = self.with_name(self.Path(new_path).name())

        with self.lock(new_path):
            if new_path.exists():
                if overwrite:
                    new_path.delete()
                else:
                    raise AttributeError(f"Target path '{new_path}' exists but overwrite is `False`.")
            self._copy_file_or_folder(new_path=new_path)

    def _copy_file_or_folder(self, new_path):
        if self.is_file():
            shutil.copy(str(self), str(new_path), follow_symlinks=False)  # Can clobber
        else:
            copy_tree(str(self), str(new_path))

    @deco_require_state(exists=True)
    def _copy_or_move(self, target_folder_path, overwrite, method):
        """ :param generalfile.Path self: """
        target_folder_path = self.Path(target_folder_path)
        if target_folder_path.is_file():
            raise NotADirectoryError("parent_path cannot be a file")

        self_parent_path = self.absolute().get_parent() if self.is_file() else self.absolute()
        if self_parent_path == target_folder_path:
            return

        if self.is_file():
            filepaths = (self,)
        else:
            filepaths = tuple(self.get_paths_in_folder())

        target_filepaths = [target_folder_path / path.absolute().relative(self_parent_path) for path in filepaths]
        if not overwrite and any([target.exists(quick=True) for target in target_filepaths]):
            raise FileExistsError("Atleast one target filepath exists, cannot copy")

        with self.lock(target_folder_path):
            target_folder_path.create_folder()
            for path, target in zip(filepaths, target_filepaths):

                if method == "copy":
                    self.__class__._copy_file_or_folder(path, target)  # Same as path._copy_file_or_folder(target)

                elif method == "move":
                    shutil.move(str(path), str(target))  # Can clobber if full target path is specified like we do

            if method == "move" and self.is_folder():
                self.delete()

    def copy_to_folder(self, target_folder_path, overwrite=False):
        """ Copy file or files inside given folder to anything except it's own parent, use `copy` for that.

            :param generalfile.Path self:
            :param target_folder_path:
            :param overwrite: """
        return self._copy_or_move(target_folder_path=target_folder_path, overwrite=overwrite, method="copy")

    def move(self, target_folder_path, overwrite=False):
        """ Move files inside given folder or file to anything except it's own parent.

            :param generalfile.Path self:
            :param target_folder_path:
            :param overwrite: """
        return self._copy_or_move(target_folder_path=target_folder_path, overwrite=overwrite, method="move")

    def is_file(self):
        """ Get whether this Path is a file.

            :param generalfile.Path self: """
        return self._path.is_file()

    def is_folder(self):
        """ Get whether this Path is a folder.

            :param generalfile.Path self: """
        return self._path.is_dir()

    def exists(self, quick=False):
        """ Get whether this Path exists.

            :param generalfile.Path self:
            :param quick: Whether to do a quick (case insensitive on windows) check. """
        if quick:
            return self._path.exists()
        else:
            try:
                path_list = self.get_paths_recursive(depth=0, include_self=True)
            except AttributeError:
                return False
            exists = False
            for foundPath in path_list:
                if foundPath == self:
                    exists = True
                elif str(foundPath).lower() == str(self).lower():
                    raise CaseSensitivityError(f"Same path with differing case not allowed: '{self}'")
            return exists

    def without_file(self):
        """ Get this path without it's name if it's a file, otherwise it returns itself.

            :param generalfile.Path self: """
        if self.is_file():
            return self.get_parent()
        else:
            return self

    @deco_require_state(is_folder=True)
    def get_paths_in_folder(self, relative=None):
        """ Get a generator containing every child Path inside this folder, relative if possible.

            :param generalfile.Path self:
            :param relative: """
        for child in self._path.iterdir():
            if relative is None:
                yield self.Path(child, parent=self)
            else:
                yield self.Path(child).relative(base=relative).set_parent(parent=self)

    @deco_require_state(quick_exists=True)
    def get_paths_recursive(self, depth=-1, include_self=False, include_files=True, include_folders=False, relative=None, filt=None):
        """ Get all paths that are next to this file or inside this folder.
            Todo: Filter for Path.get_paths_* like we have in ObjInfo.

            :param depth: Depth of -1 is limitless recursive searching. Depth of 1 searches only first level.
            :param include_self:
            :param include_files:
            :param include_folders:
            :param generalfile.Path self:
            :param relative:
            :param function filt: Optional filter with Path as arg. """
        if self.is_file():
            queued_folders = [self.get_parent()]
        elif self.is_folder():
            queued_folders = [self]
        else:
            raise AttributeError(f"Path {self} is neither file nor folder.")

        self_parts_len = len(queued_folders[0].parts())

        if include_self:
            yield self if relative is None else self.relative(base=relative)

        while queued_folders:
            for path in queued_folders[0].get_paths_in_folder():
                if filt and not filt(path):
                    continue

                if path.is_file():
                    if include_files and path != self:
                        yield path if relative is None else path.relative(base=relative)
                elif path.is_folder():
                    if include_folders:
                        yield path if relative is None else path.relative(base=relative)

                    current_depth = len(path.parts()) - self_parts_len + 1
                    if depth == -1 or current_depth < depth:
                        queued_folders.append(path)
            del queued_folders[0]

    def create_folder(self):
        """ Create folder with this Path unless it exists. 

            :param generalfile.Path self: """
        if self.exists():
            return False
        else:
            self._path.mkdir(parents=True, exist_ok=True)
            return True

    def open_folder(self):
        """ Open folder to view it manually.

            :param generalfile.Path self: """
        self.create_folder()
        os.startfile(str(self.without_file()))

    @classmethod
    @deco_cache()
    def get_cache_dir(cls):
        """ Get cache folder.

            :param generalfile.Path cls: """
        return cls.Path(appdirs.user_cache_dir())

    @classmethod
    @deco_cache()
    def get_lock_dir(cls):
        """ Get lock folder inside cache folder.

            :param generalfile.Path cls: """
        return cls.Path.get_cache_dir() / "generalfile" / "locks"

    @classmethod
    def get_working_dir(cls):
        """ Get current working folder as a new Path.
            Falls back to last seen working_dir if it doesn't exist. (Only seems to raise Error on posix)

            :param generalfile.Path cls:
            :rtype: generalfile.Path """
        # return cls.Path(pathlib.Path.cwd())

        try:
            working_dir = cls.Path(pathlib.Path.cwd())
        except FileNotFoundError as e:
            if cls._working_dir is None:
                raise e
            else:
                return cls._working_dir
        else:
            cls._working_dir = working_dir
            return working_dir

    def set_working_dir(self):
        """ Set current working folder.

            :param generalfile.Path self: """
        self.create_folder()
        self._working_dir = self.absolute()
        os.chdir(str(self._working_dir))

    @deco_preserve_working_dir
    @deco_return_if_removed(content=False)
    def delete(self, error=True):  # TODO: Add this error parameter for more methods
        """ Delete a file or folder.

            :param error:
            :param generalfile.Path self: """
        with self.lock():
            try:
                if self.is_file():
                    os.remove(str(self))
                elif self.is_folder():
                    shutil.rmtree(str(self), ignore_errors=True)
            except Exception as e:
                if error:
                    raise e

    @deco_preserve_working_dir
    @deco_return_if_removed(content=False)
    def trash(self):
        """ Trash a file or folder.

            :param generalfile.Path self: """
        with self.lock():
            send2trash(str(self))

    @deco_preserve_working_dir
    @deco_return_if_removed(content=True)
    def delete_folder_content(self):
        """ Delete a file or folder and then create an empty folder in it's place.

            :param generalfile.Path self: """
        for path in self.get_paths_in_folder():
            path.delete()

    @deco_preserve_working_dir
    @deco_return_if_removed(content=True)
    def trash_folder_content(self):
        """ Trash a file or folder and then create an empty folder in it's place.

            :param generalfile.Path self: """
        for path in self.get_paths_in_folder():
            path.trash()

    @deco_require_state(is_file=True)
    def seconds_since_creation(self):
        """ Get time in seconds since file was created.
            NOTE: Doesn't seem to update very quickly for windows (7).

            :param generalfile.Path self: """
        return time.time() - os.path.getctime(str(self))

    @deco_require_state(is_file=True)
    def seconds_since_modified(self):
        """ Get time in seconds since file was modified.

            :param generalfile.Path self: """
        return time.time() - os.path.getmtime(str(self))





    @deco_require_state(is_file=True)
    def size(self):
        """ Get size in bytes of file.

            :param generalfile.Path self: """
        return self._path.stat().st_size

    def is_identical(self, path):
        """ Get whether this file is identical to another.

            :param generalfile.Path self:
            :param path: """
        path = self.Path(path)
        self_exists = self.exists(quick=True)
        path_exists = path.exists(quick=True)

        if not self_exists or not path_exists:
            return self_exists == path_exists

        # if self.size() != path.size():
        #     return False

        with self.lock(path):
            with open(str(self), "r") as file1:
                with open(str(path), "r") as file2:
                    # print(file1.read())
                    # print(file2.read())
                    return file1.read() == file2.read()

    @deco_require_state(is_folder=True)
    def get_differing_files(self, target, exist=True, content=True, filt=None):
        """ Get list of changed files by comparing two folders.
            Todo: Tests for get_differing_files.

            :param generalfile.Path self:
            :param target:
            :param exist:
            :param content:
            :param filt: """
        target = self.Path(target)
        assert target.is_folder()

        self_paths = set(self.get_paths_recursive(relative=self, filt=filt))
        target_paths = set(target.get_paths_recursive(relative=target, filt=filt))

        diff = set()
        if exist:
            diff.update(self_paths.difference(target_paths))
        if content:
            diff.update({path for path in self_paths.intersection(target_paths) if not (self / path).is_identical(path=target / path)})
        return diff






















